from dataclasses import dataclass
from datetime import datetime, timedelta
from zoneinfo import ZoneInfo

import pytz


@dataclass
class Transition:
    utc_transition_time: datetime  # When the transition occurs, in UTC
    utc_offset: timedelta  # The new UTC offset (e.g. 2h for CET -> CEST)
    dst_offset: timedelta  # The new amount of DST in effect (e.g. 1h for CET -> CEST)
    tz_name: str  # The new timezone name (e.g. CEST)

    @classmethod
    def from_pytz_transition(cls, transition_time, transition_info):
        return cls(
            utc_transition_time=transition_time.replace(tzinfo=ZoneInfo("UTC")),
            utc_offset=transition_info[0],
            dst_offset=transition_info[1],
            tz_name=transition_info[2],
        )


class TransitionsData:
    def __init__(self, year_start: int = 1970, year_end: int = 2038):
        self.year_start = year_start
        self.year_end = year_end

        self.tz_transitions = None
        self.build_tz_transitions()

    def build_tz_transitions(self):
        self.tz_transitions = {}
        for iana_name in pytz.common_timezones:  # IANA: https://www.iana.org/time-zones
            try:
                pytz_transitions = zip(
                    pytz.timezone(iana_name)._utc_transition_times,
                    pytz.timezone(iana_name)._transition_info,
                )
            except AttributeError:  # Not all time zones have DST transitions
                continue

            # Add transitions for this time zone for the given year range
            self.tz_transitions[iana_name] = [
                Transition.from_pytz_transition(transition_time, transition_info)
                for transition_time, transition_info in pytz_transitions
                if self.year_start <= transition_time.year <= self.year_end
            ]

        # Special case: UTC has no transitions
        self.tz_transitions["UTC"] = []


if __name__ == "__main__":
    transition_data = TransitionsData()
    from pprint import pprint

    pprint(transition_data.tz_transitions["Europe/Amsterdam"])



from typing import Optional, Union
from zoneinfo import ZoneInfo

import pandas as pd

from tz_canary.exceptions import ImplausibleTimeZoneError
from tz_canary.infer import infer_time_zone
from tz_canary.transitions_data import TransitionsData


def validate_time_zone(
    dt_index: pd.DatetimeIndex,
    time_zone: Union[str, ZoneInfo] = None,
    transition_data: Optional[TransitionsData] = None,
) -> None:
    """Validate that the given time zone is plausible given DST changes in the index.

    This function can be used in one of two ways:
    1. If `dt_index` has a time zone, this function will check whether it is plausible.
    2. If `dt_index` has no time zone but the `time_zone` argument is given, this
         function will check whether that time zone is plausible.

    You cannot pass both a `time_zone` and a time-zone-aware `dt_index`.

    Args:
        dt_index: A pandas DatetimeIndex. If `dt_index` is time-zone-aware, its time
            zone will be validated. If `dt_index` time-zone-naive, the `time_zone`
            argument must be given.
        time_zone: A time zone to validate. If `time_zone` is given, `dt_index` must be
            time-zone-naive.
        transition_data: A TransitionsData object. If None, the TransitionsData will be
            built for the years spanning the given index. When inferring time zones for
            many indices, it is more efficient to build a TransitionsData object once
            and pass it to multiple calls of this function.

    Raises:
        ImplausibleTimeZoneError: If the given time zone is implausible.
        ValueError: If both `dt_index` and `time_zone` are given.
    """

    if (dt_index.tz is None) == (time_zone is None):
        raise ValueError(
            "Pass either a time-zone-aware `dt_index` or a `time_zone`, but not both. "
            "If `dt_index` is time-zone-naive, pass a `time_zone`. "
            "If `dt_index` is time-zone-aware, do not pass a `time_zone`."
        )

    given_time_zone = time_zone or dt_index.tz  # TODO - normalize to ZoneInfo
    if isinstance(given_time_zone, str):
        given_time_zone = ZoneInfo(given_time_zone)

    plausible_time_zones = infer_time_zone(dt_index, transition_data)

    if given_time_zone not in plausible_time_zones:
        raise ImplausibleTimeZoneError(
            f"The given time zone `{given_time_zone}` is not plausible for `dt_index`. "
            f"It may be one of: `{sorted([tz.key for tz in plausible_time_zones])}`."
        )



import logging
from datetime import datetime, timedelta
from typing import Optional, Set
from zoneinfo import ZoneInfo

import pandas as pd

from tz_canary.transitions_data import Transition, TransitionsData

logger = logging.getLogger(__name__)


def infer_time_zone(
    dt_index: pd.DatetimeIndex, transition_data: Optional[TransitionsData] = None
) -> Set[ZoneInfo]:
    """Infer a set of plausible time zones based on DST switches.

    Args:
        dt_index: A pandas DatetimeIndex.
        transition_data: A TransitionsData object. If None, the TransitionsData will be
            built for the years spanning the given index. When inferring time zones for
            many indices, it is more efficient to build a TransitionsData object once
            and pass it to multiple calls of this function.

    Returns:
        A set of plausible time zones for the given index.
    """

    if transition_data is None:
        transition_data = TransitionsData(
            year_start=dt_index.min().year, year_end=dt_index.max().year
        )

    # Remove time zone if set; what the index reports to be is irrelevant for inference.
    reported_tz = dt_index.tz
    if reported_tz is not None:
        logger.info(
            f"`dt_index` claims to have time zone `{reported_tz}`. "
            "Ignoring this during time zone inference."
        )
        dt_index = dt_index.tz_localize(None)

    # Find plausible time zones.
    all_time_zones = set(ZoneInfo(tz) for tz in transition_data.tz_transitions.keys())
    plausible_time_zones = set()  # all expected DST transitions occur
    implausible_time_zones = set()  # at least one expected DST transition doesn't occur

    for tz_name, transitions in transition_data.tz_transitions.items():
        # Check if each transition occurs, does not occur, or is out of range.
        transition_occurrences = [
            _check_transition_occurs(dt_index, tz_name, transition)
            for transition in transitions
        ]

        # Discard out-of-range transitions as they do not give us any information.
        transition_occurrences_within_range = [
            t for t in transition_occurrences if t is not None
        ]

        # If no transitions occur with in the range, we cannot say whether the time zone
        #   is plausible or not.
        if len(transition_occurrences_within_range) == 0:
            continue

        # If all DST changes that are expected for the time zone occur it is a plausible
        #   time zone. Otherwise, it is not.
        if all(transition_occurrences_within_range):
            plausible_time_zones.add(ZoneInfo(tz_name))
        else:
            implausible_time_zones.add(ZoneInfo(tz_name))

    # If there are no plausible time zones specifically detected for the index (its
    #   range may not cover any potential DST changes, or it may actually have no DST
    #   changes), we say all time zones that are not implausible are plausible.
    if len(plausible_time_zones) == 0:
        plausible_time_zones = all_time_zones - implausible_time_zones

    return plausible_time_zones


def _check_transition_occurs(
    dt_index_naive: pd.DatetimeIndex, candidate_tz_name: str, transition: Transition
) -> Optional[bool]:
    """Check if a DST transition occurs in the given index.

    We check three cases:
    - Whether the transition time is within the index at all
    - Whether a spring forward DST transition occurs (if applicable to the transition)
    - Whether a fall back DST transition occurs (if applicable to the transition)

    Args:
        dt_index_naive: A time-zone-naive datetime index.
        candidate_tz_name: The name of the time zone whose transition we're checking.
        transition: The transition to check.

    Returns:
        True if the transition occurs, False if it does not, None if it is outside the
            range of the index.
    """
    if dt_index_naive.tz is not None:
        raise ValueError(
            "`dt_index_naive` may not have a time zone set. "
            "You can remove it using `.tz_localize(None)`"
        )

    # If the transition time is not within the index, it cannot have occurred.
    utc_transition_time_naive = transition.utc_transition_time.replace(tzinfo=None)
    if (
        utc_transition_time_naive < dt_index_naive.min()
        or utc_transition_time_naive > dt_index_naive.max()
    ):
        return None

    # Check if the transition occurs (spring forward or fall back).
    if _check_spring_forward_occurs(dt_index_naive, transition, candidate_tz_name):
        return True
    if _check_fall_back_occurs(dt_index_naive, transition, candidate_tz_name):
        return True

    # Neither spring forward nor fall back occurred.
    return False


def _check_spring_forward_occurs(
    dt_index_naive: pd.DatetimeIndex, transition: Transition, candidate_tz_name: str
) -> bool:
    """Check if the transition from non-DST to DST occurs (spring forward).

    When converting the UTC transition time to the new time zone, the hour (to be
    precise, the amount of time corresponding to the DST offset) before it should not
    exist in the index.

    Args:
        dt_index_naive: A time-zone-naive datetime index.
        transition: The potential spring forward transition to check.
        candidate_tz_name: The name of the time zone whose transition we're checking.

    Returns:
        True if the spring forward DST transition occurs, False if it does not.
    """

    # When we spring forward, the DST offset is positive; if it is negative, a spring
    #   forward definitely did not occur.
    if transition.dst_offset <= timedelta(0):
        return False

    # First localize the UTC transition time to the new candidate time zone, then remove
    #   the time zone information, so we can find it within the naive datetime index.
    local_transition_time_naive = _localize_naive(
        utc_time=transition.utc_transition_time, tz_name=candidate_tz_name
    )

    # Find the indices in that are within the DST offset.
    idx_during_transition = dt_index_naive[
        (dt_index_naive < local_transition_time_naive)
        & (dt_index_naive > local_transition_time_naive - transition.dst_offset)
    ]

    # If there are no indices during the DST transition, a spring forward occurred. If
    #   there are, it did not.
    return len(idx_during_transition) == 0


def _check_fall_back_occurs(
    dt_index_naive: pd.DatetimeIndex, transition: Transition, candidate_tz_name: str
) -> bool:
    """Check if the transition from DST to non-DST occurs (fall back).

    When converting the UTC transition time to the new time zone, this time should exist
    in the index twice: once with the DST offset, once without.

    Args:
        dt_index_naive: A time-zone-naive datetime index.
        transition: The potential fall back transition to check.
        candidate_tz_name: The name of the time zone whose transition we're checking.

    Returns:
        True if the fall back DST transition occurs, False if it does not.
    """

    # When we fall back, the DST offset becomes zero.
    if transition.dst_offset != timedelta(0):
        return False

    # Find the indices in that are equal to the localized, naive transition time.
    idx_during_transition = dt_index_naive[
        dt_index_naive
        == _localize_naive(transition.utc_transition_time, candidate_tz_name)
    ]

    # If there are two indices at the time of the DST transition, a fall back occurred.
    #   If not, it did not.
    return len(idx_during_transition) == 2


def _localize_naive(utc_time: datetime, tz_name: str) -> datetime:
    """Localize a time-zone-aware UTC time to a time zone and make it time-zone-naive.

    Args:
        utc_time: A time-zone-aware time in UTC.
        tz_name: The name of the time zone to localize to.

    Returns:
        The naive localized time.
    """

    return utc_time.astimezone(ZoneInfo(tz_name)).replace(tzinfo=None)



from tz_canary.exceptions import ImplausibleTimeZoneError  # noqa: F401
from tz_canary.infer import infer_time_zone  # noqa: F401
from tz_canary.transitions_data import TransitionsData  # noqa: F401
from tz_canary.validate import validate_time_zone  # noqa: F401
